@@ -1,5 +1,7 @@
+from bs4 import BeautifulSoup
 from datetime import datetime
 from dexy.exceptions import UserFeedback
+from dexy.exceptions import InternalDexyProblem
 from dexy.plugin import TemplatePlugin
 from dexy.utils import levenshtein
 from dexy.version import DEXY_VERSION
@@ -11,20 +13,26 @@ import dexy.data
 import dexy.exceptions
 import dexy.plugin
 import inflection
+import inspect
+import jinja2
 import json
+import markdown
 import operator
 import os
 import pygments
 import pygments.formatters
 import re
 import time
-import markdown
+import uuid
+import xml.etree.ElementTree as ET
 
-try:
-    from bs4 import BeautifulSoup
-    BS4_AVAILABLE = True
-except ImportError:
-    BS4_AVAILABLE = False
+class Etree(TemplatePlugin):
+    """
+    Exposes element tree as ET.
+    """
+    aliases = ['etree']
+    def run(self):
+        return { 'ET' : ("The xml.etree.ElementTree module.", ET,) }
 
 class Markdown(TemplatePlugin):
     """
@@ -34,45 +42,60 @@ class Markdown(TemplatePlugin):
 
     def run(self):
         md = markdown.Markdown()
+        h = "Function which converts markdown to HTML."
         return {
-                'markdown' : md.convert,
-                'md' : md.convert
+                'markdown' : (h, md.convert),
+                'md' : (h, md.convert)
                 }
 
+class Uuid(TemplatePlugin):
+    """
+    Exposes the UUID module.
+    """
+    aliases = ['uuid']
+    def run(self):
+        return { 'uuid' : ("The Python uuid module. http://docs.python.org/2/library/uuid.html", uuid) }
+
 class Time(TemplatePlugin):
     """
     Exposes time module.
     """
     aliases = ['time']
     def run(self):
-        return {
-                'time' : time
-                }
+        return { 'time' : ("The Python time module.", time) }
 
 class Operator(TemplatePlugin):
     """
     Exposes features of the operator module.
     """
     aliases = ['operator']
+    keys = ['attrgetter', 'itemgetter']
     def run(self):
-        return {
-                'attrgetter' : operator.attrgetter,
-                'itemgetter' : operator.itemgetter
-                }
+        d = {}
+        for k in self.keys:
+            fn = getattr(operator, k)
+            d[k] = ("The %s method from Python's operator module." % k, fn)
+        return d
 
 class PrettyPrintHtml(TemplatePlugin):
     """
     Uses BeautifulSoup 4 to prettify HTML.
     """
-    @classmethod
-    def is_active(klass):
-        return BS4_AVAILABLE
+    aliases = ['bs4']
+    _settings = {
+            'no-jinja-filter' : ['BeautifulSoup']
+            }
 
-    @classmethod
-    def prettify_html(klass, html):
-        soup = BeautifulSoup(unicode(html))
+    def prettify_html(self, html):
+        soup = BeautifulSoup(unicode(html), 'html.parser')
         return soup.prettify()
 
+    def run(self):
+        return {
+            'prettify_html' : ("Pretty-print HTML using BeautifulSoup", self.prettify_html),
+            'BeautifulSoup' : ("The BeautifulSoup module.", BeautifulSoup)
+            }
+
 class LoadYaml(TemplatePlugin):
     """
     Loads YAML from a file.
@@ -86,7 +109,7 @@ class LoadYaml(TemplatePlugin):
 
     def run(self):
         return {
-                'load_yaml' : self.load_yaml
+                'load_yaml' : ("Safely load YAML from a file.", self.load_yaml,)
                 }
 
 class Debug(TemplatePlugin):
@@ -114,16 +137,15 @@ class Debug(TemplatePlugin):
 
     def run(self):
         return {
-                'debug' : self.debug,
-                'throw' : self.throw,
-                'raise' : self.throw
+                'debug' : ("A debugging method - prints content to command line stdout.", self.debug),
+                'throw' : ("A debugging utility which raises exception with argument.", self.throw),
+                'raise' : ("Alias for `throw`.", self.throw)
                 }
 
 class Bibtex(TemplatePlugin):
     """
     Produces a bibtex entry for dexy.
     """
-    @classmethod
     def run(self):
         return { 'dexy_bibtex' : dexy.commands.cite.bibtex_text() }
 
@@ -140,27 +162,32 @@ class Inflection(TemplatePlugin):
             }
 
     def run(self):
-        args = {}
-        for method in self.setting('methods'):
-            args[method] = getattr(inflection, method)
-        return args
+        return dict((method, ("The %s method from Python inflection module." % method, getattr(inflection, method),)) for method in self.setting('methods'))
 
-class JavadocToRst(TemplatePlugin):
+class StripJavadocHTML(TemplatePlugin):
     """
     Exposes javadoc2rst command which strips HTML tags from javadoc comments.
     """
+    aliases = ['stripjavadochtml']
+    _settings = {
+            'escape' : ("Escape characters.", ['\\']),
+            'remove' : ("Remove characters.", ['<p>', '<P>'])
+            }
 
-    ESCAPE = ['\\']
-    REMOVE = ['<p>', '<P>']
-
-    @classmethod
-    def javadoc2rst(klass, javadoc):
-        for x in klass.REMOVE:
-            javadoc = javadoc.replace(x, '\n')
-        for x in klass.ESCAPE:
-            javadoc = javadoc.replace(x, "\\%s" % x)
+    def strip_javadoc_html(self, javadoc):
+        for symbol in self.setting('escape'):
+            javadoc = javadoc.replace(symbol, '\n')
+        for word in self.setting('remove'):
+            javadoc = javadoc.replace(word, "\\%s" % symbol)
         return javadoc
 
+    def run(self):
+        h = "Replace escape character with newlines and remove paragraph tags."
+        return {
+                'javadoc2rst' : (h, self.strip_javadoc_html),
+                'strip_javadoc_html' : (h, self.strip_javadoc_html)
+                }
+
 class PrettyPrint(TemplatePlugin):
     """
     Exposes pprint (really pformat).
@@ -170,8 +197,8 @@ class PrettyPrint(TemplatePlugin):
     def run(self):
         import pprint
         return {
-            'pprint' : pprint.pformat,
-            'pformat' : pprint.pformat
+            'pprint' : ("Pretty prints Python objects.", pprint.pformat,),
+            'pformat' : ("Pretty prints Python objects.", pprint.pformat,)
          }
 
 class PrettyPrintJson(TemplatePlugin):
@@ -185,48 +212,171 @@ class PrettyPrintJson(TemplatePlugin):
 
     def run(self):
         return {
-            'ppjson' : self.ppjson
+            'ppjson' : ("Pretty prints valid JSON.", self.ppjson,)
          }
 
-class JinjaFilters(TemplatePlugin):
+class ReplaceJinjaFilters(TemplatePlugin):
     """
-    Custom jinja filters which can be accessed using the | fn() syntax in jinja templates.
+    Replace some jinja filters so they call unicode() first.
     """
-    @classmethod
-    def indent(klass, s, width=4, indentfirst=False):
-        """
-        Replacement for jinja's indent method to ensure unicode() is called prior to splitlines().
-        """
-        output = []
-        for i, line in enumerate(unicode(s).splitlines()):
-            if indentfirst or i > 0:
-                output.append("%s%s" % (' ' * width, line))
-            else:
-                output.append(line)
-        return "\n".join(output)
+    aliases = ['replacejinjafilters']
 
-    @classmethod
-    def head(klass, s, n=15):
+    def do_indent(self, data, width=4, indentfirst=False):
+        return jinja2.filters.do_indent(unicode(data), width, indentfirst)
+
+    def run(self):
+        return {
+                'indent' : ("Jinja's indent function.", self.do_indent)
+                }
+
+class Assertions(TemplatePlugin):
+    """
+    Allow making assertions in documents.
+    """
+    aliases = ['assertions']
+
+    def decorate_response(self, doc):
+        if hasattr(self, 'filter_instance'):
+            indicator = self.filter_instance.setting('assertion-passed-indicator')
+        else:
+            indicator = None
+
+        if indicator:
+            return unicode(doc) + indicator
+        else:
+            return doc
+
+    def do_assert_equals(self, doc, expected):
+        """
+        Assert that input equals expected value.
+        """
+        assert unicode(doc) == expected, "input text did not equal '%s'" % expected
+        return self.decorate_response(doc)
+
+    def do_assert_contains(self, doc, contains):
+        """
+        Assert that input equals expected value.
+        """
+        assert contains in unicode(doc), "input text did not contain '%s'" % contains
+        return self.decorate_response(doc)
+
+    def do_assert_does_not_contain(self, doc, shouldnt_contain):
+        """
+        Assert that input equals expected value.
+        """
+        msg = "input text contained '%s'" % shouldnt_contain
+        assert not shouldnt_contain in unicode(doc), msg
+        return self.decorate_response(doc)
+
+    def do_assert_startswith(self, doc, startswith):
+        """
+        Assert that the input starts with the specified value.
+        """
+        assert unicode(doc).startswith(startswith), "input text did not start with '%s'" % startswith
+        return self.decorate_response(doc)
+
+    def do_assert_matches(self, doc, regexp):
+        """
+        Assert that input matches the specified regular expressino.
+        """
+        assert re.match(regexp, unicode(doc)), "input text did not match regexp %s" % regexp
+        return self.decorate_response(doc)
+
+    def make_soup(self, doc):
+        return BeautifulSoup(unicode(doc))
+
+    def soup_select(self, doc, selector):
+        soup = self.make_soup(doc)
+        return soup.select(selector)
+
+    def soup_select_unique(self, doc, selector):
+        results = self.soup_select(doc, selector)
+
+        n = len(results)
+        if n == 0:
+            msg = "no results found matching selector '%s'"
+            msgargs = (selector,)
+            raise AssertionError(msg % msgargs)
+
+        elif n > 1:
+            msg = "%s results found matching selector '%s', must be unique"
+            msgargs = (n, selector,)
+            raise AssertionError(msg % msgargs)
+
+        return results[0]
+
+    def do_assert_selector_text(self, doc, selector, expected_text):
+        """
+        Asserts that the contents of CSS selector matches the expected text.
+        
+        Leading/trailing whitespace is stripped before comparison.
+        """
+        element = self.soup_select_unique(doc, selector)
+        err = "element '%s' did not contain '%s'" % (selector, expected_text)
+        assert element.get_text().strip() == expected_text, err
+
+    def run(self):
+        methods = {}
+        for name in dir(self):
+            if name.startswith("do_"):
+                method = getattr(self, name)
+                docs = inspect.getdoc(method).splitlines()[0].strip()
+                if not docs:
+                    raise InternalDexyProblem("You must define docstring for %s" % name)
+                methods[name.replace("do_", "")] = (docs, method)
+        return methods
+
+class Head(TemplatePlugin):
+    """
+    Provides a 'head' method.
+    """
+    aliases = ['head']
+
+    def head(self, text, n=15):
         """
         Returns the first n lines of input string.
         """
-        lines = unicode(s).splitlines()[0:n]
-        return "\n".join(lines)
+        return "\n".join(unicode(text).splitlines()[0:n])
+
+    def run(self):
+        return {
+                'head' : ("Returns the first n lines of input string.", self.head)
+                }
+
+class Tail(TemplatePlugin):
+    """
+    Provides a 'tail' method.
+    """
+    aliases = ['tail']
+
+    def tail(self, text, n=15):
+        """
+        Returns the last n lines of input string.
+        """
+        return "\n".join(unicode(text).splitlines()[-n:])
+
+    def run(self):
+        return {
+                'tail' : ("Returns the last n lines of inptu string.", self.tail)
+                }
 
 class RstCode(TemplatePlugin):
     """
-    Indents code 4 spaces and wraps in .. code:: directive.
+    Indents code n spaces (defaults to 4) and wraps in .. code:: directive.
     """
-    @classmethod
-    def rstcode(klass, text, language='python'):
-        output = [
-            ".. code:: %s" % language,
-            '   :class: highlight',
-            '']
+    aliases = ['rstcode']
+
+    def rstcode(self, text, n=4, language='python'):
+        output = inspect.cleandoc("""
+            .. code:: %s
+            '   :class: highlight'
+            """ % language)
+
         for line in unicode(text).splitlines():
-            output.append("    %s" % line)
-        output.append('')
-        return os.linesep.join(output)
+            output += " " * n + line
+
+        output += "\n"
+        return output
 
 class PythonDatetime(TemplatePlugin):
     """
@@ -241,13 +391,13 @@ class PythonDatetime(TemplatePlugin):
         caldates = list(cal.itermonthdates(year, month))
 
         return {
-            "datetime" : datetime,
-            "calendar" : calendar,
-            "caldates" : caldates,
-            "cal" : cal,
-            "today" : today,
-            "month" : month,
-            "year" : year
+            "datetime" : ("The Python datetime module.", datetime),
+            "calendar" : ("A Calendar instance from Python calendar module.", calendar),
+            "caldates" : ("List of calendar dates in current month.", caldates),
+            "cal" : ("Shortcut for `calendar`.", calendar),
+            "today" : ("Result of datetime.today().", today),
+            "month" : ("Current month.", month),
+            "year" : ("Current year.", year)
             }
 
 class DexyVersion(TemplatePlugin):
@@ -256,7 +406,7 @@ class DexyVersion(TemplatePlugin):
     """
     aliases = ['dexyversion']
     def run(self):
-        return { "DEXY_VERSION" : DEXY_VERSION }
+        return { "DEXY_VERSION" : ("The active dexy version. Currently %s." % DEXY_VERSION, DEXY_VERSION) }
 
 class SimpleJson(TemplatePlugin):
     """
@@ -264,7 +414,7 @@ class SimpleJson(TemplatePlugin):
     """
     aliases = ['json']
     def run(self):
-        return { 'json' : json }
+        return { 'json' : ("The Python json module.", json,) }
 
 class RegularExpressions(TemplatePlugin):
     """
@@ -272,7 +422,7 @@ class RegularExpressions(TemplatePlugin):
     """
     aliases = ['regex']
     def run(self):
-        return { 're_match' : re.match, 're_search' : re.search}
+        return { 're' : ("The Python re module.", re,), }
 
 class PythonBuiltins(TemplatePlugin):
     """
@@ -280,32 +430,26 @@ class PythonBuiltins(TemplatePlugin):
     """
     aliases = ['builtins']
     # Intended to be all builtins that make sense to run within a document.
-    PYTHON_BUILTINS = [abs, all, any, bin, bool, bytearray, callable, chr,
-        cmp, complex, dict, dir, divmod, enumerate, filter, float, format, hex,
-        id, int, isinstance, issubclass, iter, len, list, locals, long, map,
-        hasattr, max, min, oct, ord, pow, range, reduce, repr, reversed, round,
-        set, slice, sorted, str, sum, tuple, xrange, unicode, zip]
+    PYTHON_BUILTINS = [abs, all, any, basestring, bin, bool, bytearray,
+            callable, chr, cmp, complex, dict, dir, divmod, enumerate, filter,
+            float, format, hex, id, int, isinstance, issubclass, iter, len,
+            list, locals, long, map, hasattr, max, min, oct, ord, pow, range,
+            reduce, repr, reversed, round, set, slice, sorted, str, sum, tuple,
+            type, xrange, unicode, zip]
 
     def run(self):
-        return dict((f.__name__, f) for f in self.PYTHON_BUILTINS)
+        return dict((f.__name__, ("The python builtin function %s" % f.__name__, f,)) for f in self.PYTHON_BUILTINS)
 
 class PygmentsStylesheet(TemplatePlugin):
     """
-    Generates pygments stylesheets.
+    Inserts pygments style codes.
     """
     aliases = ['pygments']
 
-    # TODO figure out default fmt based on document ext
-    @classmethod
-    def highlight(klass, text, lexer_name, fmt = 'html', noclasses = False, lineanchors = 'l'):
-        if text:
-            text = unicode(text)
-            formatter_options = { "lineanchors" : lineanchors, "noclasses" : noclasses }
-            lexer = pygments.lexers.get_lexer_by_name(lexer_name)
-            formatter = pygments.formatters.get_formatter_by_name(fmt, **formatter_options)
-            return pygments.highlight(text, lexer, formatter)
+    # TODO rewrite this so it's a function rather than pre-generating all
+    # of the stylesheets. Detect document format automatically.
 
-    def run(self):
+    def generate_stylesheets(self):
         pygments_stylesheets = {}
         if hasattr(self, 'filter_instance') and self.filter_instance.doc.args.has_key('pygments'):
             formatter_args = self.filter_instance.doc.args['pygments']
@@ -324,7 +468,36 @@ class PygmentsStylesheet(TemplatePlugin):
                         ext = 'css'
                     key = "%s.%s" % (style_name, ext)
                     pygments_stylesheets[key] = style_info
-        return {'pygments' : pygments_stylesheets, 'highlight' : self.highlight }
+
+        return pygments_stylesheets
+
+    def run(self):
+        return {
+            'pygments' : (
+                "Dictionary of pygments stylesheets.",
+                self.generate_stylesheets()
+             )}
+
+class PygmentsHighlight(TemplatePlugin):
+    """
+    Provides a 'highlight' function for applying syntax highlighting.
+    """
+    aliases = ['highlight']
+    # TODO figure out default fmt based on document ext - document would need
+    # to implement a "final_ext()" method
+
+    def highlight(self, text, lexer_name, fmt='html', noclasses=False, lineanchors='l'):
+        text = unicode(text)
+        formatter_options = { "lineanchors" : lineanchors, "noclasses" : noclasses }
+        lexer = pygments.lexers.get_lexer_by_name(lexer_name)
+        formatter = pygments.formatters.get_formatter_by_name(fmt, **formatter_options)
+        return pygments.highlight(text, lexer, formatter)
+
+    def run(self):
+        return {
+                'highlight' : ("Pygments syntax highlighter.", self.highlight),
+                'pygmentize' : ("Pygments syntax highlighter.", self.highlight)
+                }
 
 class Subdirectories(TemplatePlugin):
     """
@@ -332,12 +505,15 @@ class Subdirectories(TemplatePlugin):
     """
     aliases = ['subdirectories']
     def run(self):
-        # The directory containing the document to be processed.
-        doc_dir = os.path.dirname(self.filter_instance.output_data.name)
+        if hasattr(self.filter_instance, 'output_data'):
+            # The directory containing the document to be processed.
+            doc_dir = os.path.dirname(self.filter_instance.output_data.name)
 
-        # Get a list of subdirectories under this document's directory.
-        subdirectories = [d for d in sorted(os.listdir(os.path.join(os.curdir, doc_dir))) if os.path.isdir(os.path.join(os.curdir, doc_dir, d))]
-        return {'subdirectories' : subdirectories}
+            # Get a list of subdirectories under this document's directory.
+            subdirectories = [d for d in sorted(os.listdir(os.path.join(os.curdir, doc_dir))) if os.path.isdir(os.path.join(os.curdir, doc_dir, d))]
+            return {'subdirectories' : ("List of subdirectories of this document.", subdirectories)}
+        else:
+            return {'subdirectories' : ("List of subdirectories of this document.", [])}
 
 class Variables(TemplatePlugin):
     """
@@ -348,7 +524,16 @@ class Variables(TemplatePlugin):
         variables = {}
         variables.update(self.filter_instance.setting('variables'))
         variables.update(self.filter_instance.setting('vars'))
-        return variables
+
+        formatted_variables = {}
+        for k, v in variables.iteritems():
+            if isinstance(v, tuple):
+                formatted_variables[k] = v
+            elif isinstance(v, list) and len(v) == 2:
+                formatted_variables[k] = v
+            else:
+                formatted_variables[k] = ("User-provided variable.", v)
+        return formatted_variables
 
 class Globals(TemplatePlugin):
     """
@@ -362,7 +547,7 @@ class Globals(TemplatePlugin):
         for kvpair in raw_globals.split(","):
             if "=" in kvpair:
                 k, v = kvpair.split("=")
-                env[k] = v
+                env[k] = ("Global variable %s" % k, v)
         return env
 
 class Inputs(TemplatePlugin):
@@ -375,21 +560,25 @@ class Inputs(TemplatePlugin):
         return self.filter_instance.doc.walk_input_docs()
 
     def run(self):
-        self.map_relative_refs = {}
-
         input_docs = {}
+
         for doc in self.input_tasks():
             input_docs[doc.key] = doc
 
         d = D(self.filter_instance.doc, input_docs)
 
+        if hasattr(self.filter_instance, 'output_data'):
+            output_data = self.filter_instance.output_data
+        else:
+            output_data = None
+
         return {
-            'a' : d,
-            'args' : self.filter_instance.doc.args,
-            'd' : d,
-            'f' : self.filter_instance,
-            's' : self.filter_instance.output_data,
-            'w' : self.filter_instance.doc.wrapper
+            'a' : ("Another way to reference 'd'. Deprecated.", d),
+            'args' : ("The document args.", self.filter_instance.doc.args),
+            'd' : ("The 'd' object.", d),
+            'f' : ("The filter instance for this document.", self.filter_instance),
+            's' : ("The data instance for this document.", output_data),
+            'w' : ("The wrapper for the dexy run.", self.filter_instance.doc.wrapper)
             }
 
 class D(object):
